const asyncAuto = require('async/auto');
const {chanNumber} = require('bolt07');
const {returnResult} = require('asyncjs-util');

const {channelEdgeAsChannel} = require('./../../lnd_responses');
const {isLnd} = require('./../../lnd_requests');

const edgeIsZombieErrorMessage = 'edge marked as zombie';
const edgeNotFoundErrorMessage = 'edge not found';

/** Get graph information about a channel on the network

  Requires `info:read` permission

  `inbound_base_discount_mtokens` is not supported on LND 0.17.5 and below
  `inbound_rate_discount` is not supported on LND 0.17.5 and below

  {
    id: <Standard Format Channel Id String>
    lnd: <Authenticated LND API Object>
  }

  @returns via cbk or Promise
  {
    capacity: <Maximum Tokens Number>
    id: <Standard Format Channel Id String>
    policies: [{
      [base_fee_mtokens]: <Base Fee Millitokens String>
      [cltv_delta]: <Locktime Delta Number>
      [fee_rate]: <Fees Charged in Millitokens Per Million Number>
      [inbound_base_discount_mtokens]: <Source Based Base Fee Reduction String>
      [inbound_rate_discount]: <Source Based Per Million Rate Reduction Number>
      [is_disabled]: <Channel Is Disabled Bool>
      [max_htlc_mtokens]: <Maximum HTLC Millitokens Value String>
      [min_htlc_mtokens]: <Minimum HTLC Millitokens Value String>
      public_key: <Node Public Key String>
      [updated_at]: <Edge Last Updated At ISO 8601 Date String>
    }]
    transaction_id: <Transaction Id Hex String>
    transaction_vout: <Transaction Output Index Number>
    [updated_at]: <Channel Last Updated At ISO 8601 Date String>
  }
*/
module.exports = ({id, lnd}, cbk) => {
  return new Promise((resolve, reject) => {
    return asyncAuto({
      // Check arguments
      validate: cbk => {
        try {
          chanNumber({channel: id}).number
        } catch (err) {
          return cbk([400, 'ExpectedValidChannelIdToGetChannel', {err}]);
        }

        if (!isLnd({lnd, method: 'getChanInfo', type: 'default'})) {
          return cbk([400, 'ExpectedLndToGetChannelDetails']);
        }

        return cbk();
      },

      // Get channel
      getChannel: ['validate', ({}, cbk) => {
        return lnd.default.getChanInfo({
          chan_id: chanNumber({channel: id}).number,
        },
        (err, response) => {
          if (!!err && err.details === edgeIsZombieErrorMessage) {
            return cbk([404, 'FullChannelDetailsNotFound']);
          }

          if (!!err && err.details === edgeNotFoundErrorMessage) {
            return cbk([404, 'FullChannelDetailsNotFound']);
          }

          if (!!err) {
            return cbk([503, 'UnexpectedGetChannelInfoError', {err}]);
          }

          if (!response) {
            return cbk([503, 'ExpectedGetChannelResponse']);
          }

          try {
            return cbk(null, channelEdgeAsChannel(response));
          } catch (err) {
            return cbk([503, err.message]);
          }
        });
      }],
    },
    returnResult({reject, resolve, of: 'getChannel'}, cbk));
  });
};
